"""Functions and helpers related to guessing configuration attributes."""


import os
import re
from collections import OrderedDict

from .constants import INPUT


def python_modules():
    """Determine if there are python modules in the cwd.

    Returns:
        list of python modules as strings
    """

    ignored = ["setup.py", "conftest.py"]

    py_modules = []
    for file_ in os.listdir(os.path.abspath(os.curdir)):
        if file_ in ignored or not os.path.isfile(file_):
            continue

        file_name, file_ext = os.path.splitext(file_)
        if file_ext == ".py":
            py_modules.append(file_name)

    return sorted(py_modules)


def potential_data_files(scripts):
    """Determine if there are any potential data files down from cwd.

    Tries to ignore normal things that would pop up you'd like not to include.

    Args:
        scripts: list of known script exectuables

    Returns:
        list of files as relative paths down from cwd
    """

    potential_files = []

    for root, directories, files in os.walk(os.curdir):
        if _ignored(root, is_file=False):
            continue

        for file_ in files:
            if not _ignored(file_):
                relative_path = os.path.join(root, file_)
                while relative_path[0] in (os.curdir, os.path.sep):
                    relative_path = relative_path[1:]
                if relative_path not in scripts:
                    potential_files.append(relative_path)

    return potential_files


def _ignored(file_or_dir, is_file=True, _recurse=False):
    """Return a boolean of if this file or dir is ignored.

    Args::

        file_or_dir: string file or directory name
        is_file: boolean of if this is a file
    """

    ignored_files = [
        "MANIFEST.in",
        "setup.cfg",
        ".travis.yml",
        ".coverage",
        ".coveragerc",
        ".gitignore",
        "tox.ini",
        "conftest.py",
        "pypackage.meta",
    ]

    ignored_file_patterns = [
        r".*\.pyc$",
        r".*\.py$",  # These will be found as part of a package
    ]

    ignored_dirs = [
        "__pycache__",
        "build",
        "dist",
        ".git",
        ".svn",
        ".eggs",
    ]

    ignored_dir_patterns = [
        r"docs?",
        r"examples?",
        r".*\.egg$",
        r".*\.egg-info$",
        r"tests?",
    ]

    if not is_file:
        return any([_ignored(part, _recurse=True) for
                    part in file_or_dir.split(os.path.sep)])

    file_or_dir = os.path.basename(file_or_dir)
    ignored = ignored_dirs if _recurse else ignored_files

    if file_or_dir in ignored:
        return True

    patterns = ignored_dir_patterns if _recurse else ignored_file_patterns
    for pattern in patterns:
        if re.match(pattern, file_or_dir):
            return True

    return False


def _guess_at_things(config):
    """Guesses at attributes of this package's setup configuration.

    Returns:
        OrderedDict of {attribute: guess}
    """

    guesses = OrderedDict([
        ("name", os.path.basename(os.path.realpath(os.path.curdir))),
        ("py_modules", python_modules()),
    ])

    scripts = []
    for potential in ["scripts", "bin"]:
        root = os.path.join(os.path.abspath(os.curdir), potential)
        if os.path.isdir(root):
            for file_ in os.listdir(root):
                if os.access(os.path.join(root, file_), os.X_OK):
                    scripts.append(os.path.join(potential, file_))

    guesses["scripts"] = scripts

    package_files = potential_data_files(scripts)
    if package_files:
        guesses["package_data"] = package_files

    return guesses


def perform_guesswork(config, options):
    """Look for missing attributes and take a guess at them.

    Confirm with the user that the guesses are correct after making them if
    they are in interactive mode, otherwise just display them as found.

    Args::

        config: config object to inspect/set attrributes on
        options: options object for cmd line flags
    """

    guesses = _guess_at_things(config)

    if options.interactive:
        # confirm guesses...
        print((
            "pypackage has guessed the following attributes:\n{}\n"
            "would you like to disregard any of the above guesses?"
        ).format(
            "\n".join(["{} `{}` = {!r:}".format(
                index,
                attr,
                guesses[attr]
            ) for index, attr in enumerate(guesses, 1)])
        ))
        len_g = len(guesses)
        query = 'use 1-{} or "all" to ignore (enter to accept): '.format(len_g)
        from_user = True
        while from_user and any([guesses[key] for key in guesses]):
            from_user = INPUT(query)
            try:
                from_user = int(from_user)
                assert 1 <= from_user <= len_g  # no negatives, inside len_g
            except:
                if str(from_user).strip().lower() == "all":
                    print("ignoring all guesses")
                    guesses = {}
                elif from_user:
                    print("{} is invalid.".format(from_user))
            else:
                # this is a hack, and the reason guess is an OrderedDict
                from_user = list(guesses.keys())[from_user - 1]
                guesses[from_user] = None
                print("{} will not be guessed".format(from_user))

    for atr in ("name", "scripts", "py_modules", "package_data"):
        if (not hasattr(config, atr) or options.re_probe) and guesses.get(atr):
            if atr == "package_data":  # set the name as late as possible
                setattr(config, atr, {
                    getattr(config, "name", guesses["name"]): guesses[atr]
                })
            else:
                setattr(config, atr, guesses[atr])

    if getattr(config, "package_data", None) and \
            not getattr(config, "include_package_data", None):
        config.include_package_data = True



"""Tests for the pypackage_setup function in the __init__.py."""


import sys
import mock
import pytest

import pypackage
from pypackage import pypackage_setup


@pytest.mark.parametrize("flag", ("-h", "--help"))
def test_help_is_raised(reset_sys_argv, flag):
    """Ensure additional help is shimmed inside the help msg per function."""

    sys.argv = ["py-build", flag]
    with pytest.raises(SystemExit) as error:
        pypackage_setup(additional="super important information")
    assert "super important information" in error.value.args[0]


@pytest.mark.parametrize("flag", ("-v", "--version"))
def test_version_is_raised(reset_sys_argv, flag):
    """Ensure the version info is raised with SystemExit when requested."""

    sys.argv = ["py-build", flag]
    with pytest.raises(SystemExit) as error:
        pypackage_setup()
    assert error.value.args[0] == pypackage.VERSION


@pytest.mark.parametrize("flag", ("-i", "--interactive"))
def test_interactive_setup_called(capfd, simple_module, reset_sys_argv, flag):
    """Ensure we call run_interactive_setup when requested."""

    sys.argv = ["py-build", "-N", flag]
    with mock.patch.object(pypackage, "run_interactive_setup") as inter_patch:
        pypackage_setup()

    assert inter_patch.call_count == 1

    out, err = capfd.readouterr()
    assert "~ setup.py ~" in out
    assert "from setuptools import setup" in out
    assert "setup(" in out
    assert not err


@pytest.mark.parametrize("flag", ("-R", "--reclassify"))
def test_reclassify_trigger(simple_module, reset_sys_argv, flag):
    """Ensure set_value_in_config is called for classifiers if requested."""

    sys.argv = ["py-build", "-N", flag]
    with mock.patch.object(pypackage, "set_value_in_config") as set_v_patch:
        pypackage_setup()

    assert set_v_patch.call_count == 1


if __name__ == "__main__":
    pytest.main(["-rx", "-v", "--pdb", __file__])



"""Reads static configuration metadata and creates a Python object."""


from __future__ import print_function

import os
import re
import copy
import json
import nose
import yaml
import pytest
import logging
import unittest
from pprint import pformat
from collections import OrderedDict
from setuptools import find_packages

from .runner import TestRunner
from .runner import NOSE_TEMPLATE
from .runner import PYTEST_TEMPLATE
from .runner import UNITTEST_TEMPLATE
from .constants import META_NAME
from .constants import STRING_TYPE


class UNDEF(object):
    """Used to differentiate None/0/False from an option not being defined."""

    pass


class Config(object):
    """Config object. Attributes are passed as kwargs."""

    _KEYS = OrderedDict([  # all available configuration keys: value type
        ("name", str),
        ("version", str),
        ("description", str),
        ("long_description", str),
        ("author", str),
        ("author_email", str),
        ("maintainer", str),
        ("maintainer_email", str),
        ("url", str),
        ("download_url", str),
        ("packages", list),
        ("py_modules", list),
        ("scripts", list),
        ("entry_points", {str: list}),
        ("install_requires", (str, list)),
        ("tests_require", (str, list)),
        ("classifiers", list),
        # ~~ separation of standard/extended here ~~
        ("ext_modules", list),
        ("keywords", list),
        ("platforms", list),
        ("cmdclass", {str: object}),
        ("data_files", list),
        ("package_dir", {str: str}),
        ("build_requires", (str, list)),
        ("setup_requires", (str, list)),
        ("extras_require", {str: list}),
        ("include_package_data", bool),
        ("exclude_package_data", {str: list}),
        ("package_data", {str: list}),
        ("zip_safe", bool),
        ("dependency_links", list),
        ("namespace_packages", list),
        ("test_suite", str),
        ("test_loader", str),
        ("eager_resources", list),
        ("use_2to3", bool),
        ("convert_2to3_doctests", list),
        ("use_2to3_fixtures", list),
    ])
    # our distinct keys, these do not get passed to setuptools/distutils
    _PYPACKAGE_KEYS = OrderedDict([
        ("test_runner", str),
        ("tests_dir", str),     # no function required for these two, they're
        ("runner_args", list),  # only used in the enable_test_runner function
    ])

    def __init__(self, **kwargs):
        """Builds a Config object, fills in options passed as kwargs."""

        config_as_dict = {key: UNDEF() for key in Config._KEYS}
        config_as_dict.update({key: UNDEF() for key in Config._PYPACKAGE_KEYS})
        self._defaults = site_defaults()
        config_as_dict.update(self._defaults)
        for key, value in kwargs.items():
            config_as_dict[key] = value

        extras = {}
        for key, value in config_as_dict.items():
            if not isinstance(value, UNDEF):
                if re.match(".*_requires?$", key) and key not in Config._KEYS:
                    extras[key.rpartition("_")[0]] = value
                else:
                    setattr(self, key, value)

        if extras:
            self.extras_require = extras

        # duplicate in some values
        duplicate_values = [
            ("description", "long_description"),
            ("author", "maintainer"),
            ("author_email", "maintainer_email"),
        ]
        for orig, alias in duplicate_values:
            if hasattr(self, orig):
                setattr(self, alias, getattr(self, alias, getattr(self, orig)))
                if getattr(self, orig) == getattr(self, alias):
                    self._defaults[alias] = getattr(self, alias)

        # _verify will toggle these if set
        self._configured_runner_args = False
        self._configured_tests_require = False

        # perform init-time type validation and runs feature functions
        self._verify()

    @property
    def _as_kwargs(self):
        """Builds a dict suitable for use with setuptools.setup directly."""

        kwargs = OrderedDict()

        for key in Config._KEYS:
            if hasattr(self, key):
                if key == "packages":
                    for package in self.packages:
                        if package.startswith("find_packages(") and \
                                not ";" in package and not "." in package:
                            # no ;'s or .'s here to reduce attack surface
                            try:
                                found_packages = eval(package)
                            except:
                                continue
                            else:
                                kwargs[key] = found_packages
                                break
                    else:
                        kwargs[key] = self.packages
                else:
                    kwargs[key] = getattr(self, key)

        return kwargs


    @property
    def _metadata(self):
        """Metadata unique to this config."""

        metadata = OrderedDict([(k, v) for k, v in self._as_kwargs.items()
                                if k not in self._defaults])
        # remove the cmdclass key and add in our feature keys
        metadata.pop("cmdclass", None)
        if not self._configured_tests_require:
            metadata.pop("tests_require", None)

        for attr in Config._PYPACKAGE_KEYS:
            if hasattr(self, attr):
                if attr != "runner_args" or self._configured_runner_args:
                    metadata[attr] = getattr(self, attr)
        return metadata

    def __str__(self):
        """Creates a string representation of self, resulting in a setup.py."""

        packages_str, find_needed = self._packages_string()
        cmdclass = self._cmdclass_string()
        imports = ["from setuptools import setup"]
        if find_needed:
            imports.append("from setuptools import find_packages")

        return "\n".join([
            '"""{}\'s setup.py.\n'.format(
                getattr(self, "name", "None").title()
            ),
            "This file was autogenerated by pypackage. To make changes, you",
            "should edit the {} rather than this setup.py.".format(META_NAME),
            '"""\n\n',
            "\n".join(imports),
            self._test_runner_string() or "\n",
            "setup(",
            "\n".join([
                "    {}={},".format(key, _multiline(val)) for key, val in
                self._as_kwargs.items() if key not in ("cmdclass", "packages")
            ]),
            "{}{})".format(
                "    {},\n".format(packages_str) if packages_str else "",
                "    {},\n".format(cmdclass) if cmdclass else "",
            ),
        ])

    def _test_runner_string(self):
        """Builds a string from the test_runner template, if in use."""

        if not hasattr(self, "test_runner"):
            return

        if self.test_runner == "unittest":  # extra formatting
            return UNITTEST_TEMPLATE.format(
                self=self,
                tests_dir=getattr(self, "tests_dir", "."),
            )
        elif self.test_runner == "nose":
            return NOSE_TEMPLATE.format(self=self)
        else:
            return PYTEST_TEMPLATE.format(self=self)

    def _packages_string(self):
        """Builds a string for `packages=` in the setup.py.

        Returns:
            tuple of a packages string and boolean of requiring find_packages
        """

        # defaults
        excludes = ["*.tests", "*.tests.*", "tests.*", "tests"]
        package_str = "find_packages(exclude={!r:})".format(excludes)
        find_required = True

        if hasattr(self, "packages"):
            for package in self.packages:
                if "find_packages(" in package:
                    # user supplied a find_packages() string
                    package_str = package
                    break
            else:
                # user supplied packages list
                package_str = repr(self.packages)
                find_required = False

        return "packages={}".format(package_str), find_required

    def _cmdclass_string(self):
        """Builds a string for `cmdclass=` in the setup.py."""

        cmdclass = getattr(self, "cmdclass", {})
        as_string = copy.deepcopy(cmdclass)
        if "test" in as_string:
            as_string["test"] = "PyPackageTest"  # name in template

        if as_string:
            return "cmdclass={{{}}}".format(", ".join(  # repr the keys only
                "{!r:}: {}".format(k, v) for k, v in as_string.items()
            ))

    def __repr__(self):
        return "<{name}.{cls} object at {id}, {attributes}>".format(
            name=__name__,
            cls=self.__class__.__name__,
            id=hex(id(self)),
            attributes=", ".join(
                ["{attr}: '{{self.{attr}}}'".format(attr=attr)
                 for attr in dir(self) if not attr.startswith("_")]
            ).format(self=self),
        )

    def _enable_test_runner(self):
        """Sets up the preferred testing suite.

        Returns:
            None, modifies attributes of self
        """

        if not hasattr(self, "test_runner"):
            return

        test_runner = self.test_runner.lower()
        if test_runner == "pytest":
            self._enable_pytest()
        elif test_runner.startswith("nose"):  # allow nosetests... etc
            self.test_runner = "nose"  # exact name for importing though
            self._enable_nosetest()
        else:
            self.test_runner = "unittest"
            self._enable_unittest()

        TestRunner.pypackage(self)  # XXX after runner_args are set
        self.cmdclass = {"test": TestRunner}

    def _enable_nosetest(self):
        """Do nosetest specific logic to enable it as a test runner."""

        default_args = ["-v", "-d", "--with-coverage", "--cov-report",
                        "term-missing", "--cov"]

        self._runner = nose

        # grab the user's tests_require, make sure nose is in there
        self.tests_require = getattr(self, "tests_require", None)
        if self.tests_require is None:
            self.tests_require = ["nose"]
        else:
            self._configured_tests_require = self.tests_require not in (
                ["nose"], ["nose", "nose-cov"])
            if "nose" not in self.tests_require:
                self.tests_require.append("nose")

        # configure the default or user supplied runner arguments
        arg_len = 1 + (int(hasattr(self, "tests_dir")) * 2)  # *2 b/c -w flag
        self.runner_args = getattr(self, "runner_args", None)
        if self.runner_args is None:
            self.runner_args = default_args[:2]
            if hasattr(self, "name"):
                self.runner_args.extend(default_args[2:] + [self.name])
                if "nose-cov" not in self.tests_require:
                    self.tests_require.append("nose-cov")
        elif len(self.runner_args) == len(default_args) + arg_len and \
                self.runner_args[:-arg_len] == default_args:
            # refresh runner_args in case our name has changed for coverage
            self.runner_args = default_args + [getattr(self, "name", "")]
            if "nose-cov" not in self.tests_require:
                self.tests_require.append("nose-cov")
        else:
            self._configured_runner_args = True  # include them in metadata

        # use -w to specify NOSEWHERE, or let nose find the tests itself
        if hasattr(self, "tests_dir"):
            self.runner_args.extend(["-w", self.tests_dir])

    def _enable_unittest(self):
        """Do unittest specific logic to enable it as a test runner."""

        self._runner = unittest
        self.runner_args = getattr(self, "runner_args", None)
        if self.runner_args is None:
            self.runner_args = []
        else:
            self._configured_runner_args = True

    def _enable_pytest(self):
        """Do pytest specific logic to enable it as a test runner."""

        default_args = ["-v", "-rx", "--cov-report", "term-missing", "--cov"]

        self._runner = pytest

        # grab the user's tests_require, make sure pytest is in there
        self.tests_require = getattr(self, "tests_require", None)
        if self.tests_require is None:
            self.tests_require = ["pytest"]
        else:
            self._configured_tests_require = self.tests_require not in (
                ["pytest"], ["pytest", "pytest-cov"])
            if "pytest" not in self.tests_require:
                self.tests_require.append("pytest")

        # configure the default or user supplied runner arguments
        arg_len = 1 + int(hasattr(self, "tests_dir"))  # name + dir if supplied
        self.runner_args = getattr(self, "runner_args", None)
        if self.runner_args is None:
            self.runner_args = default_args[:2]
            if hasattr(self, "name"):
                self.runner_args.extend(default_args[2:] + [self.name])
                if "pytest-cov" not in self.tests_require:
                    self.tests_require.append("pytest-cov")
        elif len(self.runner_args) == len(default_args) + arg_len and \
                self.runner_args[:-arg_len] == default_args:
            # refresh runner_args in case our name has changed for coverage
            self.runner_args = default_args + [getattr(self, "name", "")]
            if "pytest-cov" not in self.tests_require:
                self.tests_require.append("pytest-cov")
        else:
            self._configured_runner_args = True  # include them in metadata

        # tack the tests dir on the end, or let pytest find them
        if hasattr(self, "tests_dir"):
            self.runner_args.append(self.tests_dir)

    def _verify(self):
        """Ensures self attributes conform to their type declarations."""

        self._enable_test_runner()

        for key, type_ in list(Config._KEYS.items()) + list(
                Config._PYPACKAGE_KEYS.items()):
            if hasattr(self, key):
                self._verify_key(key, type_)

    def _verify_key(self, key, type_):
        """Verify that key is of type type_.

        Raises:
            TypeError if they key cannot be coerced into type_
        """

        if isinstance(type_, dict) and isinstance(getattr(self, key), dict):
            setattr(self, key, ensure_dict(getattr(self, key), type_))
        elif isinstance(type_, dict):
            raise TypeError("{} should be a dict, not {}!".format(
                key,
                type(getattr(self, key)).__name__,
            ))
        elif type_ is list and isinstance(getattr(self, key), list):
            setattr(self, key, ensure_list(getattr(self, key)))
        elif type_ is list:
            setattr(self, key, [getattr(self, key)])
        elif not isinstance(getattr(self, key), type_):
            if isinstance(type_, tuple):  # multiple acceptable values
                for type__ in type_:
                    if type__ is list:
                        setattr(self, key, [getattr(self, key)])
                        break
                    else:
                        try:
                            setattr(self, key, type__(getattr(self, key)))
                            break
                        except:
                            pass
                else:
                    raise TypeError("{} should be a {} or {}, not {}!".format(
                        key,
                        ", ".join([t.__name__ for t in type_[:-1]]),
                        type_[-1].__name__,
                        type(getattr(self, key)).__name__,
                    ))
            else:
                try:
                    setattr(self, key, type_(getattr(self, key)))
                except:
                    raise TypeError("{} should be a {}, not {}!".format(
                        key, type_.__name__, type(getattr(self, key)).__name__,
                    ))


def _multiline(value, indent=4):
    """Return value as a multiline (pretty) string, with indent."""

    val = pformat(value)
    if "\n" in val:
        value_as_lines = val.splitlines()

        indented = [
            "{}".format(value_as_lines[0][0]),  # rip the leading bracket
            "{}{}".format(" " * (indent * 2), value_as_lines[0][1:].strip()),
        ]
        for line in value_as_lines[1:-1]:
            indented.append("{}{}".format(" " * (indent * 2), line.strip()))
        indented.extend([
            "{}{}".format(" " * (indent * 2), value_as_lines[-1][:-1].strip()),
            "{}{}".format(" " * indent, value_as_lines[-1][-1]),
        ])
        val = "\n".join(indented)
    return val


def ensure_list(list_to_verify):
    """Ensures that all items in the list are of string type.

    Also guarentees that all items in the list are unique. Maintains order.
    """

    string_list = []

    def string_list_append(item):
        if item not in string_list:
            string_list.append(item)

    for item in list_to_verify:
        if isinstance(item, STRING_TYPE):
            string_list_append(item)
        else:
            string_list_append(str(item))

    return string_list


def ensure_dict(dict_to_verify, dict_types):
    """Goes through dict_to_verify and tries to coerce into dict_types.

    Args::

        dict_to_verify: a dict with keys and values to verify
        dict_types: a dict with a single key/value pair of the expected types

    Returns:
        a new dictionary, coerced into the correct types
    """

    key_type = list(dict_types.keys())[0]
    value_type = list(dict_types.values())[0]
    return_dict = {}
    for key, value in dict_to_verify.items():
        if not isinstance(key, key_type):
            key = key_type(key)  # this will raise on error...
        if not isinstance(value, value_type):
            value = value_type(value)  # this will also raise on error...
        return_dict[key] = value
    return return_dict


def site_defaults():
    """Look for site default metadata to mixin onto any missing values."""

    filename = os.path.join(os.path.expanduser("~"), ".pypackage")
    if os.path.isfile(filename):
        return json_maybe_commented(filename) or {}
    else:
        logging.warn("Site defaults requested but not found at %s", filename)
        return {}


def get_config(path=None):
    """Looks in path or os.cwd for a static metadata file and builds a config.

    Merges missing values from the DefaultConfig object

    Args:
        path: optional string path to search inside, or None for cwd

    Returns:
        Config object to build/install with or None on error
    """

    if path is None:
        path = os.path.abspath(os.path.curdir)

    pyjson = os.path.join(path, META_NAME)
    if os.path.isfile(pyjson):
        return Config(**json_maybe_commented(pyjson))
    else:
        logging.warn("Using site defaults, no %s found in %s", META_NAME, path)
        return Config(**site_defaults())


def reduce_json_unicode(json_obj):
    """Stupidly dumps JSON into YAML to remove what unicode it can."""

    return yaml.safe_load(json.dumps(json_obj))


def json_maybe_commented(filename, remove_comments=False):
    """Tries twice to load the filename as JSON.

    The second time, if needed, it will remove lines starting with `#`.
    """

    if remove_comments:
        cleaned_lines = []
        with open(filename, "r") as openfile:
            for line in openfile.readlines():
                if not re.match("\s*#", line):  # leading whitespace then #
                    cleaned_lines.append(line)
        try:
            return reduce_json_unicode(json.loads("".join(cleaned_lines)))
        except Exception as error:
            logging.error("Error reading json from %s: %r", filename, error)
            return {}
    else:
        try:
            with open(filename, "r") as openfile:
                return reduce_json_unicode(json.loads(openfile.read()))
        except Exception as error:
            logging.debug("Commented JSON? Recursing because: %r", error)
            return json_maybe_commented(filename, remove_comments=True)



"""Tests for pypackages' config object and helper functions."""


import mock
import codecs
import pytest
import tempfile

from pypackage import config
from pypackage.config import Config


def test_requires_into_extras():
    """Any x_requires param should be converted to an extras_require param."""

    conf = Config(foo_requires=["bar"], bar_requires=["foo"],
                  baz_require=["foobar"])
    assert conf.extras_require == {
        "foo": ["bar"],
        "bar": ["foo"],
        "baz": ["foobar"],
    }


def test_malformed_packages_fallback():
    """Invalid values using find_packages should fallback to their string."""

    garbage = "find_packages(!@!^%&)%*_!()$*!*^!%&*(!)$@_!*)"
    conf = Config(packages=[garbage])
    assert conf.packages == [garbage]
    assert conf._as_kwargs.get("packages") == [garbage]
    assert conf._packages_string() == ("packages={}".format(garbage), True)


@pytest.mark.parametrize("runner", ("unittest", "nose", "pytest"))
def test_runner_args_only_when_set(runner):
    """runner_args should only be in the metadata when they're non-default."""

    conf = Config(test_runner=runner)
    assert "runner_args" not in conf._metadata

    conf2 = Config(test_runner=runner, runner_args=["-vv", "--pdb"])
    assert conf2._metadata.get("runner_args") == ["-vv", "--pdb"]


@pytest.mark.parametrize("runner", ("nose", "pytest"))
def test_runner_args_with_tests_dir(runner):
    """If pytest or nose is used, the tests dir should be the last runner arg.

    When unittest is used, the tests_dir is injected into UNITTEST_TEMPLATE.
    """

    conf = Config(test_runner=runner, tests_dir="somewhere")
    assert conf.runner_args[-1] == "somewhere"


def test_runner_string__unittest():
    """Ensure the correct template and formatting is used for unittest."""

    conf = Config(
        test_runner="unittest is the default when provided nonsense",
        tests_dir="my_non_std_tests",
    )
    template_str = conf._test_runner_string()

    assert "import unittest" in template_str
    assert "my_non_std_tests" in template_str


@pytest.mark.parametrize("runner", ("nose", "pytest"))
def test_runner_string(runner):
    """Ensure the correct template is used for nose and pytest."""

    conf = Config(test_runner=runner)
    template_str = conf._test_runner_string()
    assert "import {}".format(runner) in template_str


def test_cmdclass_string():
    """Ensure the cmdclass string outputting is correct."""

    conf = Config(cmdclass={"foo": "MyFooClass", "test": "gets overridden"})
    cmdcls_str = conf._cmdclass_string()
    assert cmdcls_str.startswith("cmdclass={")
    assert "'foo': MyFooClass" in cmdcls_str
    assert "'test': PyPackageTest" in cmdcls_str


def test_repr():
    """Config should have some sort of repr that's not a setup.py."""

    conf = Config()
    assert "Config object at {}".format(hex(id(conf))) in repr(conf)


@pytest.mark.parametrize("runner", ("nose", "pytest"))
def test_additional_tests_require(runner):
    """If there are provided tests_require, they should be mixed in."""

    # we get runner-cov here becuase we provide a name
    conf = Config(name="test", test_runner=runner, tests_require=["my_thing"])
    assert conf.tests_require == ["my_thing", runner, "{}-cov".format(runner)]
    assert conf.runner_args[-1] == "test"
    # change our name here, we should get our new name in runner_args
    conf.name = "changed"
    # also we can reset tests_require and they should re-populate
    conf.tests_require = ["my_thing"]
    conf._verify()
    assert conf.runner_args[-1] == "changed"


def test_verify_key__dict_failure():
    """Do not try to coerce something that should be a dict into one."""

    conf = Config()
    conf.foo = "not a dict"
    with pytest.raises(TypeError) as error:
        conf._verify_key("foo", {})
    assert error.value.args[0] == "foo should be a dict, not str!"


def test_verify_key__list_object():
    """If an attribute should be a list, it should be listed."""

    conf = Config()
    conf.foo = "not a list"
    conf._verify_key("foo", list)
    assert conf.foo == ["not a list"]


def test_verify_key__multiple__allowed():
    """There are some keys where multiple types are allowed."""

    conf = Config()
    conf.foo = 3.14
    conf._verify_key("foo", (list, str))
    assert conf.foo == [3.14]

    conf.foo = 3.14
    conf._verify_key("foo", (str, list))  # assert order matters
    assert conf.foo == "3.14"


def test_verify_key__multiple__fallthrough():
    """If the first coerce fails, try the second."""

    conf = Config()
    conf.bar = mock.MagicMock(spec=False)
    conf.bar.__float__ = mock.Mock(side_effect=ValueError)
    conf.bar.__str__ = mock.Mock(return_value="mock str")
    conf._verify_key("bar", (float, str))
    assert conf.bar == "mock str"


def test_verify_key__multiple__failure():
    """When unable to coerce to any type, raise TypeError."""

    conf = Config()
    conf.foo = mock.MagicMock(spec=False)
    conf.foo.__float__ = mock.Mock(side_effect=ValueError)
    conf.foo.__str__ = mock.Mock(side_effect=ValueError)
    with pytest.raises(TypeError) as err:
        conf._verify_key("foo", (float, str))
    assert err.value.args[0] == "foo should be a float or str, not MagicMock!"


def test_verify_key__failure_coerce():
    """Should try to coerce values into their types when not list or dict."""

    conf = Config()
    conf.foo = 3.14
    conf._verify_key("foo", str)
    assert conf.foo == "3.14"


def test_verify_key__failure():
    """If unable to coerce into the expected type, raise TypeError."""

    conf = Config()
    conf.foo = "something"
    with pytest.raises(TypeError) as error:
        conf._verify_key("foo", float)
    assert error.value.args[0] == "foo should be a float, not str!"


def test_verify_key__str_list_values():
    """If the type is list, the items inside should be strings."""

    conf = Config()
    conf.foo = ["bar", 3.14, "baz"]
    conf._verify_key("foo", list)
    assert conf.foo == ["bar", "3.14", "baz"]


def test_verify_key__dict_types():
    """If a dict is provided with types, the key/values should be coerced."""

    conf = Config()
    conf.foo = {3.14: "500.123"}
    conf._verify_key("foo", {str: float})
    assert conf.foo == {"3.14": 500.123}


def test_multiline():
    """Ensure the proper formatting for multiline pprint formats."""

    blurb = "really long string that will require multiple lines"
    blurber = lambda x : "{}-{}".format(blurb, x)
    value = [blurber(i) for i in range(5)]
    assert config._multiline(value) == """[
        'really long string that will require multiple lines-0',
        'really long string that will require multiple lines-1',
        'really long string that will require multiple lines-2',
        'really long string that will require multiple lines-3',
        'really long string that will require multiple lines-4'
    ]"""


def test_json_loading__failure():
    """Ensure the error logged and empty return value on invalid json file."""

    filename = tempfile.mktemp()
    with open(filename, "w") as openfile:
        openfile.write("# yeah comments are ok\n{but invalid json: [is not]}")

    with mock.patch.object(config.logging, "error") as patched_err_log:
        assert config.json_maybe_commented(filename, True) == {}

    assert patched_err_log.call_args[0][0] == "Error reading json from %s: %r"
    assert patched_err_log.call_args[0][1] == filename
    assert isinstance(patched_err_log.call_args[0][2], ValueError)
    assert patched_err_log.call_count == 1


if __name__ == "__main__":
    pytest.main(["-rx", "-v", "--pdb", __file__])



"""Constants used in pypackage."""


import os
import sys
import pkg_resources


if sys.version_info > (3,):  # pragma: no cover
    INPUT = input
    STRING_TYPE = str
    CHECKMARK = chr(0x2713)
else:  # pragma: no cover
    INPUT = raw_input
    STRING_TYPE = basestring
    CHECKMARK = unichr(0x2713)


VERSION = "{} {}".format(
    os.path.basename(sys.argv[0]),
    pkg_resources.get_distribution("pypackage").version,
)


META_NAME = "pypackage.meta"


HELP = """Pypackage is an interactive and automatic Python packaging tool.

Package metadata comes from three sources (overridden in order):

    1) Site defaults, located in $HOME/.pypackage. These are applied to every
       package you make, and should be constrained to generic attributes such
       as author, emails, etc.
    2) Guesswork, done internally in pypackage. It will make guesses for what
       python modules or packages you have, your package name, scripts and/or
       static data to include.
    3) Package metadata. Read from the current directory's {}, this
       JSON-ish (full line comments are allowed/ignored) file can contain keys
       for every/any kwarg that either setuptools or distutils can handle.

Pypackage uses setuptools to perform any installation or building of packages.

{{additional}}

Usage:
    {} [OPTIONS]

Options:
    -a --all                Same as -e/--extended, display all config options
    -e --extended           Consider all options for interactive configuration
    -h --help               Show this help message and exit
    -i --interactive        Enter interactive configuration mode
    -m --metadata           Only update the package metadata; build a setup.py
    -N --no-guess           Do not perform any guessing of attributes
    -p --reprobe            Re-guess all attributes, ignore package metadata
    -r --reset --rebuild    Rebuild package metadata, ignore current settings
    -R --reclassify         (re)Enter the classifiers selection screen
    -s --setup              Only build a setup.py; don't run it
    -v --version            Show version information and exit\
""".format(
    META_NAME,
    os.path.basename(sys.argv[0]),
)



"""Choose python classifiers with a curses frontend."""


from __future__ import unicode_literals

import os
import curses
from collections import namedtuple

from .constants import VERSION
from .constants import CHECKMARK


class BoxSelector(object):  # pragma: no cover
    """Originally designed for accman.py.

    Display options build from a list of strings in a (unix) terminal.
    The user can browser though the textboxes and select one with enter.

    Used in pypackage to display the python trove classifiers in a somewhat
    logical/easy to navigate way. The unfortunate part is that this uses
    curses to display this to the user. Ideally a cross-platform solution can
    be found to replace this class.

    Known issues:
        curses incorrectly handles unicode, might look like crap, YMMV
        curses uses (y,x) for coordinates because fuck your logic
        curses support on winderps is sketchy/non-existant
    """

    # Author: Nikolai Tschacher
    # Date: 02.06.2013
    # adapted for use in pypackage by Adam Talsma in May 2015

    def __init__(self, classifier, screen, choices=None, current=0):
        """Create a BoxSelector object.

        Args:
            classifier: the Classifier root to find choices inside of
            screen: the curses screen object
            choices: a list of values in the classifier that are selected
            current: integer index of classifiers/values to start on
        """

        self.stdscr = screen

        choices = choices or []
        self.current_selected = current

        selections = []
        if classifier.name != "__root__":
            selections.append("..")

        for group in classifier.classifiers:
            selections.append("[+] {}".format(group.name))

        for value in classifier.values:
            selections.append(" {}  {}".format(
                CHECKMARK if value in choices else " ",
                value,
            ))

        # Element parameters. Change them here.
        self.TEXTBOX_WIDTH = max(79, max([len(i) for i in selections]) + 2)
        self.TEXTBOX_HEIGHT = 3

        if classifier.name == "__root__":
            selections.append("Done".center(self.TEXTBOX_WIDTH - 4, " "))

        self.L = selections

        self.PAD_WIDTH = 600
        self.PAD_HEIGHT = 10000

    def pick(self):
        """Runs the user selection proccess, returns their choice index."""

        self._init_curses()
        self._create_pad()

        windows = self._make_textboxes()
        picked = self._select_textbox(windows)

        self._end_curses()

        return picked

    def _init_curses(self):
        """Initializes the curses appliation."""

        # turn off automatic echoing of keys to the screen
        curses.noecho()
        # Enable non-blocking mode. Keys are read without hitting enter
        curses.cbreak()
        # Disable the mouse cursor.
        curses.curs_set(0)
        self.stdscr.keypad(1)
        # Enable colorous output.
        curses.start_color()
        curses.init_pair(1, curses.COLOR_BLACK, curses.COLOR_GREEN)
        curses.init_pair(2, curses.COLOR_WHITE, curses.COLOR_BLACK)
        self.stdscr.bkgd(curses.color_pair(2))
        self.stdscr.refresh()

    def _end_curses(self):
        """Terminates the curses application."""

        curses.nocbreak()
        self.stdscr.keypad(0)
        curses.echo()
        curses.endwin()

    def _create_pad(self):
        """Creates a big self.pad to place the textboxes in."""

        self.pad = curses.newpad(self.PAD_HEIGHT, self.PAD_WIDTH)
        self.pad.box()

    def _make_textboxes(self):
        """Build the textboxes in the center of the pad."""

        # Get the actual screensize.
        maxy, maxx = self.stdscr.getmaxyx()

        banner = "{} -- choose python trove classifiers".format(VERSION)
        self.stdscr.addstr(0, maxx // 2 - len(banner) // 2, banner)

        windows = []
        i = 2
        for item in self.L:
            pad = self.pad.derwin(
                self.TEXTBOX_HEIGHT,
                self.TEXTBOX_WIDTH,
                i,
                self.PAD_WIDTH // 2 - self.TEXTBOX_WIDTH // 2,
            )
            pad.box()
            try:
                pad.addstr(1, 2, item)
            except UnicodeEncodeError:
                # curses has fucked unicode support
                item = item.replace(CHECKMARK, "*")
                pad.addstr(1, 2, item)
            windows.append(pad)
            i += self.TEXTBOX_HEIGHT

        return windows

    def _center_view(self, window):
        """Centers and aligns the view according to the window argument given.

        Returns:
            the (y, x) coordinates of the centered window
        """

        # The refresh() and noutrefresh() methods of a self.pad require 6 args
        # to specify the part of self.pad to be displayed and the location on
        # the screen to be used for the display. The arguments are pminrow,
        # pmincol, sminrow, smincol, smaxrow, smaxcol; the p arguments refer
        # to the top left corner of the self.pad region to be displayed and the
        # s arguments define a clipping box on the screen within which the
        # self.pad region is to be displayed.
        cy, cx = window.getbegyx()
        maxy, maxx = self.stdscr.getmaxyx()
        self.pad.refresh(cy, cx, 1, maxx // 2 - self.TEXTBOX_WIDTH // 2,
                         maxy - 1, maxx - 1)
        return (cy, cx)

    def _select_textbox(self, windows):
        """Handles keypresses and user selection."""

        # See at the root textbox.
        topy, topx = self._center_view(windows[0])

        last = self.current_selected - 1
        top_textbox = windows[0]

        while True:
            # Highligth the selected one, the last selected textbox should
            # become normal again.
            windows[self.current_selected].bkgd(curses.color_pair(1))
            windows[last].bkgd(curses.color_pair(2))

            # While the textbox can be displayed on the page with the current
            # top_textbox, don't alter the view. When this becomes impossible,
            # center the view to last displayable textbox on the previous view.
            maxy, maxx = self.stdscr.getmaxyx()
            cy, cx = windows[self.current_selected].getbegyx()

            # The current window is to far down. Switch the top textbox.
            if ((topy + maxy - self.TEXTBOX_HEIGHT) <= cy):
                top_textbox = windows[self.current_selected]

            # The current window is to far up. There is a better way though...
            if topy >= cy + self.TEXTBOX_HEIGHT:
                top_textbox = windows[self.current_selected]

            if last != self.current_selected:
                last = self.current_selected

            topy, topx = self._center_view(top_textbox)

            c = self.stdscr.getch()

            # Vim like KEY_UP/KEY_DOWN with j(DOWN) and k(UP).
            if c in (106, curses.KEY_DOWN):  # 106 == j
                if self.current_selected >= len(windows) - 1:
                    self.current_selected = 0  # wrap around.
                else:
                    self.current_selected += 1
            elif c in (107, curses.KEY_UP):  # 107 == k
                if self.current_selected <= 0:
                    self.current_selected = len(windows) - 1  # wrap around.
                else:
                    self.current_selected -= 1
            elif c == 113:  # 113 = q == Quit without selecting.
                break
            # At hitting enter, return the index of the selected list element.
            elif c == curses.KEY_ENTER or c == 10:
                return int(self.current_selected)
            elif c == 27:  # esc or alt, try to determine which
                self.stdscr.nodelay(True)
                n_seq = self.stdscr.getch()
                self.stdscr.nodelay(False)
                if n_seq == -1:
                    # Escape was pressed, check if the top option has .. in it
                    # import pdb; pdb.set_trace()
                    if ".." in str(windows[0].instr(1, 0)):
                        return 0  # backs up a level
                    else:
                        break  # exits


Classifier = namedtuple("Classifier", ("name", "values", "classifiers"))


def _ensure_chain(top_level, sub_categories):
    """Ensure a chain of Classifiers from top_level through sub_categories."""

    def _chain_in(level, item):
        for sub_class in level.classifiers:
            if sub_class.name == item:
                return sub_class
        else:
            new_sub = Classifier(item, [], [])
            level.classifiers.append(new_sub)
            return new_sub

    for sub_cat in sub_categories:
        top_level = _chain_in(top_level, sub_cat)

    return top_level


def read_classifiers():
    """Reads the trove file and returns a Classifier representing all."""

    # TODO: get file from https://pypi.python.org/pypi?:action=list_classifiers
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
              "classifiers")) as openc:
        classifiers = [c.strip() for c in openc.read().splitlines() if c]

    all_classifiers = []

    def _get_classifier(categories):
        """Find or create the classifier for categories."""

        top_level = categories[0]
        sub_categories = categories[1:]
        for classifier in all_classifiers:
            if classifier.name == top_level:
                top_level = classifier
                break
        else:
            top_level = Classifier(top_level, [], [])
            all_classifiers.append(top_level)

        return _ensure_chain(top_level, sub_categories)

    for clsifier in classifiers:
        _get_classifier(clsifier.split(" :: ")[:-1]).values.append(clsifier)

    return Classifier("__root__", [], all_classifiers)


def back_it_up(current_level, all_classifiers, recursive=False):
    """Returns the classifier up a level from current."""

    for classifier in all_classifiers.classifiers:
        if current_level in classifier.classifiers:
            return classifier

    for classifier in all_classifiers.classifiers:
        attempt = back_it_up(current_level, classifier, True)
        if attempt:
            return attempt

    if not recursive:
        return all_classifiers


def choose_classifiers(config):
    """Get some user input for the classifiers they'd like to use.

    Returns:
        list of valid classifier names
    """

    all_classifiers = read_classifiers()
    root_classifier = all_classifiers
    old_delay = os.getenv("ESCDELAY")
    os.environ["ESCDELAY"] = "25"  # the default delay is a full second...
    screen = curses.initscr()
    choices = getattr(config, "classifiers", [])
    choice = BoxSelector(root_classifier, screen, choices).pick()
    while choice is not None:
        init = 0
        if choice == 0 and root_classifier.name != "__root__":
            root_classifier = back_it_up(root_classifier, all_classifiers)
        elif choice == 9 and root_classifier.name == "__root__":
            break   # the "done" box from the top level
        elif choice > len(root_classifier.classifiers):
            choice_index = (choice - len(root_classifier.classifiers) -
                            int(root_classifier.name != "__root__"))
            choice_as_str = root_classifier.values[choice_index]
            if choice_as_str not in choices:
                choices.append(choice_as_str)
            else:
                choices.remove(choice_as_str)
            init = choice
        else:
            choice_index = choice - int(root_classifier.name != "__root__")
            root_classifier = root_classifier.classifiers[choice_index]

        choice = BoxSelector(root_classifier, screen, choices, init).pick()

    if old_delay:
        os.environ["ESCDELAY"] = old_delay
    else:
        os.environ.pop("ESCDELAY")

    return choices



"""Parses flags passed on the command line.

If you're looking for the entry points, they're in commands.py, this only
deals with user options.
"""


import sys


def flags(*args):
    """Returns a boolean of if any of the flags are selected.

    -single character args can be compressed, ie; -e -g == -eg
    """

    cmd_line_flags = []
    for flag in sys.argv:
        if flag.startswith("-") and "--" not in flag and len(flag) > 2:
            cmd_line_flags.extend(["-{}".format(f) for f in flag[1:]])
        else:
            cmd_line_flags.append(flag)
    return any([arg in cmd_line_flags for arg in args])


def get_options():
    """Search through argv real quick and lazy like for some flags.

    Returns:
        configuration object with boolean attributes:
        re_classify, re_config, interactive, setup, metadata, extended,
        re_probe, no_guess, help
    """

    class Options(object):
        def __init__(self):
            self.re_classify = flags("-R", "--reclassify")
            self.re_config = flags("-r", "--reset", "--rebuild")
            self.interactive = flags("-i", "--interactive")
            self.setup = flags("-s", "--setup")
            self.metadata = flags("-m", "--metadata")
            self.extended = flags("-e", "--extended", "--all", "-a")
            self.re_probe = flags("-p", "--reprobe")
            self.no_guess = flags("-N", "--no-guess")
            self.help = flags("-h", "--help")
            self.version = flags("-v", "--version")

    return Options()



"""Tests for pypackages' guessing features."""


import os
import sys
import mock
import pytest
from collections import OrderedDict

from pypackage import guessing
from pypackage.config import Config
from pypackage.cmdline import get_options


def test_ignored_data_files(with_data):
    """Ensure that files in ignored directories are ignored."""

    root, pkg_root = with_data
    os.mkdir(os.path.join(pkg_root, "build"))
    with open(os.path.join(pkg_root, "build", "data"), "w") as opendata:
        opendata.write("some data")

    data_files = guessing.potential_data_files([])
    expected = [os.path.join(os.path.basename(pkg_root), "data", "data_1")]
    assert data_files == expected


def test_perform_guesswork(capfd, reset_sys_argv, move_home_pypackage):
    """Ensure the user can deselect guesses when using interactive."""

    conf = Config()
    conf.name = "previously existing"
    sys.argv = ["py-build", "-i"]
    guesses = OrderedDict([
        ("name", "some name"),
        ("py_modules", "some_thing"),
        ("scripts", ["bin/something"]),
        ("package_data", ["thing/data/file_1"]),
    ])

    with mock.patch.object(guessing, "_guess_at_things", return_value=guesses):
        with mock.patch.object(guessing, "INPUT",
                               side_effect=iter(["1", "-3", "2", ""])):
            guessing.perform_guesswork(conf, get_options())

    assert conf.name == "previously existing"
    assert not hasattr(conf, "py_modules")
    assert conf.package_data == {"previously existing": ["thing/data/file_1"]}
    assert conf.scripts == ["bin/something"]

    out, err = capfd.readouterr()
    assert "name will not be guessed" in out
    assert "py_modules will not be guessed" in out

    assert not err


def test_perform_guesswork__ignore(capfd, reset_sys_argv, move_home_pypackage):
    """If the user responds with 'all', ignore all guesses."""

    conf = Config()
    conf.name = "previously existing"
    sys.argv = ["py-build", "-i"]
    guesses = OrderedDict([
        ("name", "some name"),
        ("py_modules", "some_thing"),
        ("scripts", ["bin/something"]),
        ("package_data", {"some name": ["thing/data/file_1"]}),
    ])

    with mock.patch.object(guessing, "_guess_at_things", return_value=guesses):
        with mock.patch.object(guessing, "INPUT", return_value="all"):
            guessing.perform_guesswork(conf, get_options())

    assert conf.name == "previously existing"
    assert not hasattr(conf, "py_modules")
    assert not hasattr(conf, "package_data")
    assert not hasattr(conf, "scripts")

    out, err = capfd.readouterr()
    assert "ignoring all guesses" in out
    assert not err


if __name__ == "__main__":
    pytest.main(["-rx", "-v", "--pdb", __file__])



"""Tests for the configure functions in pypackage."""


import sys
import mock
import pytest

from pypackage import configure
from pypackage.config import Config
from pypackage.cmdline import get_options


def test_ask_for_value():
    """Ensure we correctly prompt the user."""

    with mock.patch.object(configure, "INPUT", return_value="ok") as inp_patch:
        assert configure.ask_for_value("foo") == "ok"
    inp_patch.assert_called_once_with("Please enter the value for foo: ")


def test_ask_for_value__empty_str_is_none():
    """If the user hits enter without input, None should be returned."""

    with mock.patch.object(configure, "INPUT", return_value=""):
        assert configure.ask_for_value("foo") is None


def test_ask_for_value__interrupt_quits():
    """SystemExit should be raised if the user sends ^C."""

    with mock.patch.object(configure, "INPUT", side_effect=KeyboardInterrupt):
        with pytest.raises(SystemExit) as error:
            configure.ask_for_value("foo")
    assert error.value.args[0] == "Cancelled on user request"


def test_value_in_config__classifiers(capfd):
    """The key 'classifiers' should be handled."""

    conf = Config()
    with mock.patch.object(configure, "handle_classifiers", return_value="ok"):
        configure.set_value_in_config("classifiers", conf, conf._KEYS)
    out, err = capfd.readouterr()
    assert out == "classifiers set to: ['ok']\n"
    assert not err


def test_value_in_config__not_classifiers(capfd):
    """Any other key should be the same."""

    conf = Config()
    with mock.patch.object(configure, "ask_for_value", return_value="Frank"):
        configure.set_value_in_config("author", conf, conf._KEYS)
    out, err = capfd.readouterr()
    assert out == "author set to: 'Frank'\n"
    assert not err


def test_value_in_config__invalid_try_again():
    """If the return from the user is invalid, they should be re-prompted."""

    conf = Config()
    user_input_patch = mock.patch.object(
        configure,
        "ask_for_value",
        side_effect=iter([
            "{'what': 12.3",     # malformed input
            "{'what': 1.2}",     # valid input, but not for the spec
            "{'what': ['ok']}",  # valid input
        ])
    )
    with user_input_patch:
        configure.set_value_in_config("entry_points", conf, conf._KEYS)

    assert conf.entry_points == {"what": ["ok"]}


@pytest.mark.parametrize("value", (None, False, 0, ""))
def test_coerce__no_value(value):
    """If the value fails a truthyness test, it should be returned."""

    assert configure.coerce_to_expected(value, "foo", str) == value


@pytest.mark.parametrize("value, expected", [
    ("None", False),
    ("something", True),
    ("1", True),
    ("0", False),
    ("false", True),
    ("False", False),
])
def test_coerce__to_bool(value, expected):
    """Ensure we are properly coercing to boolean."""

    assert configure.coerce_to_expected(value, "foo", bool) is expected


@pytest.mark.parametrize("value, expected", [(15.16, "15.16"), (True, "True")])
def test_coerce__to_str(value, expected):
    """Ensure we coerce to str."""

    assert configure.coerce_to_expected(value, "foo", str) == expected


def test_coerce_to_list_sorts():
    """Ensure coerce to list returns a sorted list if it's already a list."""

    assert configure.coerce_to_list(["foo", "bar", "baz", "potato"]) == [
        "bar", "baz", "foo", "potato"
    ]


@pytest.mark.parametrize("value", ",;:| ")
def test_corce_to_list_splits(value):
    """Ensure words joined with any of the value chars are split by coerce."""

    phrase = "In the beginning were the Words".split()
    assert configure.coerce_to_list(value.join(phrase)) == [
        "In",
        "Words",
        "beginning",
        "the",
        "the",
        "were",
    ]


def test_standard_attributes(reset_sys_argv, move_home_pypackage):
    """Ensure the standard attribute set."""

    conf = Config()
    expected_attrs = list(conf._KEYS.keys())[:17]
    conf.name = "foobar"
    conf.classifiers = ["fake classifier"]
    expected_attrs.remove("name")
    expected_attrs.remove("classifiers")
    attrs = configure.standard_attributes(conf, get_options())
    assert attrs == expected_attrs


def test_standard_attributes__re_config(reset_sys_argv):
    """If reconfig is set, all standard attributes should be unconfigured."""

    conf = Config()
    conf.name = "something"
    sys.argv = ["py-build", "-r"]
    attrs = configure.standard_attributes(conf, get_options())
    expected = list(conf._KEYS.keys())[:17]
    assert attrs == expected


def test_standard_attributes__re_classify(reset_sys_argv):
    """If reclassify is set, classifiers should be in the unconfigured set."""

    conf = Config()
    conf.classifiers = ["fake things"]
    sys.argv = ["py-build", "-R"]
    attrs = configure.standard_attributes(conf, get_options())
    assert "classifiers" in attrs


def test_feature_attributes(reset_sys_argv, move_home_pypackage):
    """If we have default runner args they should appear unconfigured."""

    conf = Config()
    conf.runner_args = ["fake", "args"]
    conf._configured_runner_args = False
    attrs = configure.feature_attributes(conf, get_options())
    expected = list(conf._PYPACKAGE_KEYS.keys())
    assert attrs == expected


def test_feature_attributes__re_config(reset_sys_argv):
    """When --rebuild is used, all features should appear unconfigured."""

    conf = Config()
    conf.test_runner = "pytest"
    sys.argv = ["py-build", "--rebuild"]
    attrs = configure.feature_attributes(conf, get_options())
    expected = list(conf._PYPACKAGE_KEYS.keys())
    assert attrs == expected


def test_extended_attributes(reset_sys_argv, move_home_pypackage):
    """Extended attributes should return any unset keys past 17."""

    conf = Config()
    expected = list(conf._KEYS.keys())[17:]
    conf.use_2to3 = True
    expected.remove("use_2to3")
    attrs = configure.extended_attributes(conf, get_options())
    assert attrs == expected


def test_extended_attributes__re_config(reset_sys_argv):
    """If --rebuild is used, all extended attributes should be unconfigured."""

    conf = Config()
    conf.use_2to3 = True
    sys.argv = ["py-build", "-r"]
    attrs = configure.extended_attributes(conf, get_options())
    expected = list(conf._KEYS.keys())[17:]
    assert attrs == expected


@pytest.mark.parametrize(
    "flags, banner",
    [
        ("-e", "Starting interactive build..."),
        ("-re", "Reconfiguring..."),
    ],
    ids=("extended/normal", "re-build")
)
def test_interactive_setup(capfd, reset_sys_argv, move_home_pypackage,
                           flags, banner):
    """Ensure the calls made and feedback during the interactive setup."""

    conf = Config()
    sys.argv = ["py-build", flags]
    opts = get_options()

    standard_patch = mock.patch.object(
        configure,
        "standard_attributes",
        return_value=["standard"],
    )
    feature_patch = mock.patch.object(
        configure,
        "feature_attributes",
        return_value=["feature"],
    )
    extended_patch = mock.patch.object(
        configure,
        "extended_attributes",
        return_value=["extended"],
    )
    set_value_patch = mock.patch.object(configure, "set_value_in_config")

    with standard_patch:
        with feature_patch:
            with extended_patch:
                with set_value_patch as patched_set_value:
                    assert configure.run_interactive_setup(conf, opts) == conf

    expected_calls = [
        mock.call.Call("standard", conf, conf._KEYS),
        mock.call.Call("feature", conf, conf._PYPACKAGE_KEYS),
        mock.call.Call("extended", conf, conf._KEYS),
    ]
    assert patched_set_value.mock_calls == expected_calls

    out, err = capfd.readouterr()
    assert banner in out
    assert "~ Standard Attributes ~" in out
    assert "~ Pypackage Features ~" in out
    assert "~ Extended Attributes ~" in out
    assert not err


if __name__ == "__main__":
    pytest.main(["-rx", "-vv", "--pdb", __file__])
